from functools import wraps
import inspect
from textwrap import dedent

# Default field names
_lat = 'lat'
_lon = 'lon'
_disp = 'displacement'
_dist = 'distance'
_time = 'time'
_speed = 'speed'

# Field info for docstrings
_fields = {
  _lat: {
    'name': _lat,
    'full_name': 'latitude coordinates',
    'units': 'degrees N (-90, 90)'
  },
  _lon: {
    'name': _lon,
    'full_name': 'longitude coordinates',
    'units': 'degrees E (-180, 180)'
  },
  _disp: {
    'name': _disp,
    'full_name': 'point-to-point displacements',
    'units': 'meters'
  },
  _dist: {
    'name': _dist,
    'full_name': 'cumulative distances',
    'units': 'meters'
  },
  _time: {
    'name': _time,
    'full_name': 'cumulative time from start',
    'units': 'seconds'
  },
  _speed: {
    'name': _speed,
    'full_name': 'speed',
    'units': 'meters per second'
  },
}


def series_fn(decorated):
  argspec = inspect.getfullargspec(decorated)
  kwds = getattr(argspec, 'args')[1:]
  kwd_defaults = getattr(argspec, 'defaults')

  @doc(kwds[:-1], kwds[-1])
  @wraps(decorated)
  def wrapped(self, **kwargs):
    input_cols = [
      kwargs.get(kwd, default)
      for kwd, default in zip(kwds[:-1], kwd_defaults[:-1])
    ]

    self._validate(*[lbl for lbl in input_cols if lbl is not None])

    response = decorated(self, *input_cols).rename(kwargs.get(kwds[-1], kwd_defaults[-1]))

    return response

  return wrapped


def docsub(*docstrings, **params):
  """A decorator that performs string substitution on a docstring template.

  Args:
    *docstrings (str or callable): The string / docstring / docstring template
      to be appended in order after default docstring of decorated.
    **params: The strings which would be used to format the docstring
      templates.

  """
  def decorator(decorated):

    docstring_components = []

    if decorated.__doc__:
      docstring_components.append(dedent(decorated.__doc__))
      # decorated.__doc__ = decorated.__doc__.format(**params)

    for docstring in docstrings:
      if hasattr(docstring, '_docstring_components'):
        docstring_components.extend(
          docstring._docstring_components  # type: ignore[union-attr]
        )
      elif isinstance(docstring, str) or docstring.__doc__:
        docstring_components.append(docstring)

    decorated.__doc__ = ''.join(
      [
        component.format(**params)
        if isinstance(component, str)
        else dedent(component.__doc__ or '')
        for component in docstring_components
      ]
    )

    # Save the template for future use.
    decorated._docstring_components = docstring_components

    return decorated

  return decorator


def doc(output_field):
  def decorator(func):

    argspec = inspect.getfullargspec(func)
    args = argspec.args
    defs = argspec.defaults or ()
    nargs = len(args)
    ndefs = len(defs)
    # print(argspec)
    kwds = args[nargs-ndefs:]

    inputs = args[:nargs-ndefs]
    inputs_opt = kwds # consider checking for "None"

    # collecting docstring and docstring templates
    docstring_components = []

    description_string = f'Calculate {_fields[output_field]["full_name"]} from '
    if _lat in inputs:
      description_string += 'GPS coordinates'
    else:
      description_string += ', '.join(
        [_fields[input]['full_name'] for input in inputs]
      )

    docstring_components.append(description_string + '.\n\n')

    if func.__doc__:
      docstring_components.append(dedent(func.__doc__))

    if len(inputs) > 0:
      docstring_components.append('Args:\n')
      input_param_component = (
        '  {{pre_param}}{name} ({{klass_in}}): {{pre_param_desc}}{full_name} '
        'along the route in {units}. '
        'Must be numeric dtype.{{post_param_desc}}\n'
        # 'Defaults to "{name}".\n'
      )
      # param_components = [
      docstring_components.extend([
        input_param_component.format(**_fields[input])
        for input in inputs
      ])
      # ]

    opt_input_param_component = (
      '  {{pre_param}}{name} ({{klass_in}}): {{pre_param_desc}}{full_name} '
      'along the route in {units}. '
      'Must be numeric dtype. Default {dft}.\n'
      # 'Defaults to "{name}".\n'
    )
    for kwd, dft in zip(kwds, defs):
      if kwd in _fields:
        docstring_components.append(
          opt_input_param_component.format(
            **_fields[kwd],
            dft=dft,
          )
        )
      else:
        # IDK???
        pass

    docstring_components.append(
      'Returns:\n'
      '  {{klass_out}}: {{pre_return_desc}}{full_name} along the route '
      'in {units}.'.format(**_fields[output_field])
    )

    # formatting templates and concatenating docstring
    func.__doc__ = ''.join(docstring_components)
    func._docstring_components = docstring_components

    return func
  
  return decorator



# -*- coding: utf-8 -*-
import inspect
import math
import warnings

import numpy as np
import pandas as pd

from . import scalar
from ._decorators import doc, docsub


EARTH_RADIUS_METERS = scalar.EARTH_RADIUS_METERS

_saveme = """
`Numpy docstring ref`_

`Numpy docstring example`_

`Google docstring example`_

`Sphinx todos`_  

Todo: 
  * Update these assumpts!
  * Figure out `Sphinx todos`_!

.. _PEP 484:
  https://www.python.org/dev/peps/pep-0484/
    
.. _Numpy docstring ref:
  https://numpydoc.readthedocs.io/en/latest/format.html

.. _Numpy docstring example:
  https://sphinxcontrib-napoleon.readthedocs.io/en/latest/example_numpy.html

.. _Google docstring example:
  https://sphinxcontrib-napoleon.readthedocs.io/en/latest/example_google.html

.. _Sphinx todos:
  https://www.sphinx-doc.org/en/master/usage/extensions/todo.html

"""

_docstring_params = dict(
  klass_in='pandas.Series',
  pre_param='',
  pre_param_desc='',
  post_param_desc='',
  pre_return_desc='',
)

def _clean_series(series):
  if series.isna().all():
    raise ValueError(
      'Cannot infer fill values when there are no values at all.'
    )

  return series.fillna(method='bfill').fillna(method='ffill')


@docsub(
  """
Assumptions:
  * Earth is a perfect sphere, with radius = 
    ``pandas_x.scalar.EARTH_RADIUS_METERS``.
  * The point-to-point distances are sufficiently short (so that
    latitude distortion and curvature have limited effects).
    
""",
  **_docstring_params,
  klass_out='pandas.Series',
)
@doc('displacement')
def ds_from_xy(lat, lon):
  """
  The chosen scheme: displacement at [i] represents the distance from [i-1] 
  to [i]. 

  """

  # Convert to radians.
  lat = lat * math.pi / 180
  lon = lon * math.pi / 180

  # Project the spherical coordinates onto a plane.
  dx = EARTH_RADIUS_METERS * np.cos(lat) * lon.diff()
  dy = EARTH_RADIUS_METERS * lat.diff()

  # Calculate point-to-point displacements on the plane.
  ds = (dx ** 2 + dy ** 2) ** 0.5

  # Remove the NaN value in the first position.
  return ds.fillna(0.)


@docsub(
  **_docstring_params,
  klass_out='pandas.Series',
)
@doc('displacement')
def ds_from_s(distance):
  """
  The chosen scheme: displacement at [i] represents the distance from [i-1] 
  to [i].

  """
  # Should this assume the index at position 0 is 0, or should this
  # assume the only NaN is at position 0? Assumpts either way...
  # Also could accomplish this behavior with bfill. Hmm.
  return distance.diff().fillna(distance[0])


@docsub(
  **_docstring_params,
  klass_out='pandas.Series',
)
@doc('distance')
def s_from_ds(displacement):
  """
  The chosen scheme: displacement at [i] represents the distance from [i-1] 
  to [i]. This scheme means converting displacements to cumulative distances
  does not require any extrapolation.

  """
  return displacement.cumsum()


@docsub(
  **_docstring_params,
  klass_out='pandas.Series',
)
@doc('displacement')
def s_from_v(speed, time=None):
  """
  The chosen scheme: speed at [i] represents the distance from [i] to [i+1].
  This means distance.diff() and time.diff() are shifted by one index from 
  speed. I have chosen to extrapolate the position at the first index by 
  assuming we start at a cumulative distance of 0.

  """
  if time is None:
    time = pd.Series([i for i in range(len(speed))])    
    
  # Should this assume the index at position 0 is 0, or should this
  # assume the only NaN is at position 0? Assumpts either way...
  return (speed.shift(1) * time.diff()).cumsum().fillna(0)


@docsub(
  **_docstring_params,
  klass_out='pandas.Series',
)
@doc('speed')
def v_from_s(distance, time=None):
  """
  The chosen scheme: speed at [i] represents the distance from [i] to [i+1].
  This means distance.diff() and time.diff() are shifted by one index from speed.
  I have chosen to extrapolate the speed at the final position by ffill.

  """
  if time is None:
    time = pd.Series([i for i in range(len(distance))])

  return (distance.diff() / time.diff()).shift(-1).fillna(method='ffill')


@docsub(
  **_docstring_params,
  klass_out='pandas.Series',
)
@doc('speed')
def v_from_ds(displacement, time=None):
  """
  The chosen scheme: displacement at [i] represents the distance from
  [i-1] to [i], while speed at [i] represents the distance from [i] to [i+1].
  This means displacements and time.diff() are shifted by one index from speed.
  I have chosen to extrapolate the speed at the final position by ffill.

  """
  if time is None:
    time = pd.Series([i for i in range(len(displacement))])

  return (displacement / time.diff()).shift(-1).fillna(method='ffill')


@docsub(
  **_docstring_params,
  # klass_out='list',
)
def reduced_point_index(lat, lon, min_dist=15.0):
  """
  Eliminates gps points that are too close together.

  No matter how far apart the points, always returns the start and
  end points.

  Originally developed in my mapmatching package; an old version
  still exists there.

  Args:
    {pre_param}lat ({klass_in}): {pre_param_desc}latitude values along the path.{post_param_desc}
    {pre_param}lon ({klass_in}): {pre_param_desc}longitude values along the path.{post_param_desc}
    {pre_param}min_dist (float): The minimum distance (meters) between the
      resulting downsampled GPS coordinates. Default 15.

  Returns:
    list: A boolean array that can be used to subsample a ``pandas.Series``
    corresponding to this GPS trace, based on this minimum distance scheme.
  """
  assert len(lat) == len(lon)

  n_records = len(lat)

  # Infer values where there are NaNs.
  lat = _clean_series(lat)
  lon = _clean_series(lon)
  
  # Initialize the boolean index, assuming only that we want to sample
  # the first and last points.
  ix_reduced = [(i == 0 or i == n_records - 1) for i in range(n_records)]
  record_num_cur = 0
  record_num_next = 1  
  while record_num_next < n_records - 1:
  
    dist_p2p = scalar.cartesian(
      lon[record_num_cur], 
      lat[record_num_cur],
      lon[record_num_next],
      lat[record_num_next],
    )

    # Check if the next record location is far enough away from the
    # current one to be included in the downsampled data.
    # Note: this works because python does not use pointers.
    # Otherwise I think the two record_num vars would be the same var.
    if dist_p2p >= min_dist:
      # The next point is sufficiently far from the current point, so
      # change the corresponding boolean index point to indicate a 
      # sample should be made there.
      ix_reduced[record_num_next] = True
      
      # Set the valid record number as the new base point.
      record_num_cur = record_num_next
    
    # Keep looking for sufficiently far-away points, starting with
    # the following one.
    record_num_next += 1

  return ix_reduced
  # return pd.Series(ix_reduced)


# -*- coding: utf-8 -*-

import os
from setuptools import setup


def read(rel_path):
  """Read a file so python does not have to import it.
  
  Inspired by (taken from) pip's `setup.py`.
  """
  here = os.path.abspath(os.path.dirname(__file__))
  # intentionally *not* adding an encoding option to open, See:
  #   https://github.com/pypa/virtualenv/issues/201#issuecomment-3145690
  with open(os.path.join(here, rel_path), 'r') as fp:
    return fp.read()


def get_version(rel_path):
  """Manually read through a file to retrieve its `__version__`.
  
  Inspired by (taken from) pip's `setup.py`.
  """
  for line in read(rel_path).splitlines():
    if line.startswith('__version__'):
      # __version__ = '0.0.1'
      delim = "'" if "'" in line else '"'
      return line.split(delim)[1]
  raise RuntimeError('Unable to find version string.')


with open('README.md') as f:
  readme = f.read()

with open('LICENSE') as f:
  license = f.read()

pkg_name = 'pandas_x'

version = get_version(f'{pkg_name}/__init__.py')

setup(
  name='pandas-x',
  version=version,
  description='GPS/position calculation accessor for pandas DataFrames.',
  long_description=readme,
  long_description_content_type='text/markdown',
  author='Aaron Schroeder',
  author_email='aaron@trailzealot.com',
  install_requires = [
    'numpy',
    'pandas',
  ],
  url='https://github.com/aaron-schroeder/pandas-x',
  project_urls={
    'Documentation': 'https://pandas-x.readthedocs.io/en/v{version}/',
  },
  license='MIT',
  # license_files=('LICENSE',),
  packages=[pkg_name],
  classifiers=[
    'License :: OSI Approved :: MIT License',
    'Intended Audience :: Developers',
    'Programming Language :: Python :: 3',
    'Programming Language :: Python :: 3.7',
  ]
)




# -*- coding: utf-8 -*-
from functools import wraps
import inspect
import math
import warnings

import numpy as np
from numpy.testing._private.utils import decorate_methods
import pandas as pd

from . import algorithms as algs
from ._decorators import (
  series_fn, docsub,
  _fields, _lat, _lon, _dist, _disp, _speed, _time
)


@pd.api.extensions.register_dataframe_accessor('pos')
class PositionAccessor:
  """Custom accessor for a ``pandas.DataFrame`` containing data along a route.

  Each row of the DataFrame should correspond to a record during an
  activity or a point along a route. Each column represents a data stream.

  The ``.pos`` accessor allows functions to work on the underlying data by
  specifying the columns to use in the calculations.

  """

  def __init__(self, record_df):
    """
    Args:
      record_df (pandas.DataFrame): Each row represents a point from
        a route or activity.

    """
    # self._validate(pandas_obj)
    self._obj = record_df

  def _validate(self, *col_labels):
    """Validate multiple columns' data for use in numeric functions."""
    # verify each column is the correct dtype
    for col in col_labels:
      if not pd.api.types.is_numeric_dtype(self._obj[col]):
        raise AttributeError(f'Column "{col}" must be numeric dtype')

  @classmethod
  def _add_series_method(cls, decorated):
    """Take a function with Series inputs and make it a method of this class.

    Args:
      decorated (function): The function to transform into a method of this
        class. 
        
    The resulting method takes kwargs only. Where the original function
    accepts pandas.Series arguments, the method kwargs accept the labels
    of DataFrame columns. Everywhere else, the method accepts the original
    function's positional and keyword arguments as kwargs.

    This class method infers which arguments in the original function expect
    pandas.Series inputs. Any positional args are assumed to refer to 
    pandas.Series, as well as any kwargs that share a name with a field defined
    in the ``pandas_x._decorators._fields`` dict. Any kwargs that do not
    exist in this dict are preserved as-is.
    """
    
    # Get info on the wrapped function's args to adapt them into
    # the modified method's expected kwargs.
    argspec = inspect.getfullargspec(decorated)
    args = argspec.args
    defs = argspec.defaults or ()
    nargs = len(args)
    ndefs = len(defs)
    kwds = args[nargs-ndefs:]

    # Positional args -> kwargs for column labels
    df_data_req = args[:nargs-ndefs]

    # kwargs with recognized field names -> kwargs for column labels
    df_data_opt = [kwd for kwd in kwds if kwd in _fields]

    # kwargs that aren't recognized field names -> regular ol' kwargs
    kwarg_params_scalar = [kwd for kwd  in kwds if kwd not in _fields]

    @docsub(
      decorated,
      # f"""See {decorated.__name__}.""", # include full module name
      klass_in='scalar',
      klass_out='pandas.Series',
      pre_param='**',
      pre_param_desc='column label in the record DataFrame containing ',
      post_param_desc=' If a label is not provided, '
        'the parameter name itself is used.',
      pre_return_desc='',
    )
    # @wraps(decorated)
    def wrapped(self, **kwargs):
      
      # If any column name kwarg corresponding to a positional arg in the
      # wrapped function isn't passed to the method, assume  the column
      # name is the wrapped function's positional arg's parameter name.
      cols_req = [
        kwargs.get(param_name, param_name)
        for param_name in df_data_req
      ]

      # If the wrapped function is looking for optional Series data, only
      # pass data to it if the function is called with a column name
      # (ie don't assume the method user wants to pass data to the
      # optional kwarg).
      cols_opt = {
        kw_param_name: kwargs.get(kw_param_name) for kw_param_name in df_data_opt 
        if kwargs.get(kw_param_name) is not None
      }

      self._validate(*cols_req, *cols_opt.values())

      # Pass required Series, optional Series, and scalar kwargs to
      # the decorated function
      decorated_series_args = [self._obj[col] for col in cols_req]
      decorated_series_args_opt = {
        kw_param_name: self._obj[col] 
        for kw_param_name, col in cols_opt.items()
      }
      decorated_scalar_kwargs = {
        kwd: kwargs.get(kwd)
        for kwd in kwarg_params_scalar
        if kwargs.get(kwd) is not None
      }
      response = decorated(
        *decorated_series_args,
        **decorated_series_args_opt,
        **decorated_scalar_kwargs
      )
      
      if isinstance(response, pd.Series):
        # Strip it of its name, as it is likely an artifact of retrieving
        # a Series from the record DataFrame.
        return response.rename()  # .rename(kwargs.get(kwds[-1], kwd_defaults[-1]))

      return response

    # me *trying* to get the method to behave like it was typed into the class
    wrapped.__name__ = decorated.__name__
    wrapped.__qualname__ = cls.__name__ + '.' + decorated.__name__
    # nargs = decorated.__code__.co_argcount
    # ndefs = len(decorated.__defaults__ or ())
    # narg = nargs - ndefs
    # print(decorated.__code__.co_varnames[:nargs])
    # print(decorated.__defaults__)
    # print(inspect.signature(decorated))
    # print(inspect.signature(wrapped))
    # print(wrapped.__code__.co_varnames)
    
    setattr(cls, decorated.__name__, wrapped)

for function in [
  algs.s_from_ds,
  algs.ds_from_s,
  algs.s_from_v,
  algs.v_from_s,
  algs.v_from_ds,
  algs.ds_from_xy,
  algs.reduced_point_index,
]:
  PositionAccessor._add_series_method(function)
